# -*- coding: utf-8 -*-

'''
clodss: hashmap data-structure
'''

from .common import keyexists


def _mapexists(instance, db, hkey, create=True):
    'checks if a hashmap exists, optionally creates one if not'
    tables = [f'{hkey}﹁hash']
    columns = 'key TEXT UNIQUE, value TEXT'
    return keyexists('hash', tables, instance, db,
                     hkey, create, columns=columns)


def hset(instance, hkey, key, val):
    'https://redis.io/commands/hset'
    db = instance.router.connection(hkey)
    _mapexists(instance, db, hkey)
    try:
        db.execute(f'INSERT OR REPLACE INTO `{hkey}﹁hash` VALUES(?, ?)',
                   (key, val))
        db.commit()
        return 1
    finally:
        db.close()


def hget(instance, hkey, key):
    'https://redis.io/commands/hget'
    db = instance.router.connection(hkey)
    _mapexists(instance, db, hkey)
    try:
        val = db.execute(
            f'SELECT value FROM `{hkey}﹁hash` WHERE key=?', (key,)).fetchone()
        return None if val is None else val[0]
    finally:
        db.close()


def hdel(instance, hkey, key):
    'https://redis.io/commands/hdel'
    db = instance.router.connection(hkey)
    if not _mapexists(instance, db, hkey, create=False):
        return 0
    try:
        cur = db.cursor()
        cur.execute(f'DELETE FROM `{hkey}﹁hash` WHERE key=?', (key,))
        count = cur.rowcount
        cur.close()
        db.commit()
        return count
    finally:
        db.close()


def hkeys(instance, hkey):
    'https://redis.io/commands/hkeys'
    db = instance.router.connection(hkey)
    if not _mapexists(instance, db, hkey, create=False):
        return []
    try:
        return [record[0]
                for record in db.execute(f'SELECT key FROM `{hkey}﹁hash`')]
    finally:
        db.close()


def hvalues(instance, hkey):
    'https://redis.io/commands/hvalues'
    db = instance.router.connection(hkey)
    if not _mapexists(instance, db, hkey, create=False):
        return []
    try:
        return [record[0]
                for record in db.execute(f'SELECT value FROM `{hkey}﹁hash`')]
    finally:
        db.close()


def hgetall(instance, hkey):
    'https://redis.io/commands/hgetall'
    db = instance.router.connection(hkey)
    if not _mapexists(instance, db, hkey, create=False):
        return {}
    try:
        return {record[0]: record[1] for record in
                db.execute(f'SELECT key, value FROM `{hkey}﹁hash`')}
    finally:
        db.close()


def hmset(instance, hkey, mapping):
    'https://redis.io/commands/hmset'
    db = instance.router.connection(hkey)
    _mapexists(instance, db, hkey)
    try:
        db.executemany(f'INSERT OR REPLACE INTO `{hkey}﹁hash` VALUES(?, ?)',
                       mapping.items())
        db.commit()
    finally:
        db.close()


def hmget(instance, hkey, *keys):
    'https://redis.io/commands/hmget'
    db = instance.router.connection(hkey)
    _mapexists(instance, db, hkey)
    try:
        values = [db.execute(f'SELECT value FROM `{hkey}﹁hash` WHERE key=?', (
            key,)).fetchone() for key in keys]
        return [None if record is None else record[0] for record in values]
    finally:
        db.close()



# -*- coding: utf-8 -*-

'''
clodss is a data-structures on-disk store with an API largly compatible with
redis. The goal is to develop a store with the simplicity of the redis API
which scales beyond memory capacity, allows harnessing multi-core processors,
and does not burden accesses with network latency.
'''

import logging
import time
import os
import __main__
from ilock import ILock
from .router import Router
from . import hashmaps
from . import lists
from . import keys
from .common import ProblematicSymbols


def wrapmethod(method, stats=None):
    '''
    - guards all clodss methods with a key-scoped lock
    - performs sanity checks on key
    - enures the key has not expired
    '''
    def wrapper(*args, **kwargs):
        globalmethod = method.__name__ in ('keys', 'scan', 'flushdb')

        if globalmethod:
            key = '﹁'
        else:
            if len(args) < 2:
                raise TypeError('too few parameters, `key` is required')
            instance = args[0]
            key = args[1]

            if '﹁' in key:
                raise ValueError('`key` contains invalid character(s)')
            key = ProblematicSymbols.remove(key)

        if stats is not None:
            t1 = time.perf_counter()
        with ILock(f'clodss-{key}', timeout=5):
            if not globalmethod:
                instance.checkexpired(key, enforce=True)
            result = method(*args, **kwargs)
        if stats is not None:
            t = time.perf_counter() - t1
            avg, n = stats.get(method.__name__, (0, 0))
            avg = (avg * n + t) / (n + 1)
            stats[method.__name__] = (avg, n + 1)
        return result
    wrapper.__name__ = method.__name__
    return wrapper


class StrictRedis:
    'main clodss class'
    def __init__(
            self, dbpath: str = None, db: int = 0, spread_factor: int = 2,
            decode_responses: bool = False, benchmark: bool = True) -> None:
        if dbpath is None:
            d = os.path.dirname(__main__.__file__)
            if d == '':
                d = '.'
            dbpath = os.path.realpath(f'{d}/./clodss-data')
        logging.info('clodss db path: %s', dbpath)
        self.decode = decode_responses
        dbpath = os.path.join(dbpath, '%02d' % db)
        os.makedirs(dbpath, exist_ok=True)
        self._dbpath = dbpath
        self.router = Router(dbpath, spread_factor)
        self.knownkeys = {}
        self.keystoexpire = {}
        self._stats = {} if benchmark else None

        modules = [keys, lists, hashmaps]
        for module in modules:
            for attr in dir(module):
                if attr.startswith('_'):
                    continue
                method = getattr(module, attr)
                if not callable(method):
                    continue
                if attr == 'sēt':
                    # `set` is a reserved keyword
                    attr = 'set'
                setattr(StrictRedis, attr, wrapmethod(method, self._stats))

    def stats(self):
        'gets benchmarking statistics'
        return self._stats

    def dbpath(self):
        'gets base path for database files'
        return self.dbpath

    def checkexpired(self, key, enforce=False):
        'checks if a key expired and removes it if so'
        db = self.router.connection(key)
        try:
            exp = [self.keystoexpire.get(key)]
            if not exp[0]:
                query = 'SELECT time FROM `﹁expiredkeys﹁` WHERE key=?'
                exp = db.execute(query, (key,)).fetchone()
            if exp:
                t = exp[0]
                now = time.time()
                if now > t:
                    if not enforce:
                        return True
                    ekey = key.replace('"', '""')
                    tables = db.execute(
                        'SELECT name FROM sqlite_master WHERE '
                        f'type="table" AND name LIKE "{ekey}%"').fetchall()
                    queries = ';\n'.join([f'DROP TABLE `{table[0]}`'
                                          for table in tables])
                    db.executescript(queries)
                    db.commit()
                    if key in self.knownkeys:
                        del self.knownkeys[key]

                    query = 'DELETE FROM `﹁expiredkeys﹁` WHERE key=?'
                    db.execute(query, (key,))
                    db.commit()
                    return True
                return 'scheduled'
            return False
        finally:
            db.close()

    def reset(self):
        'clears a database and all cached information'
        self.router.reset()
        self.knownkeys = {}
        self.keystoexpire = {}



# -*- coding: utf-8 -*-

'''
clodss: keys-related functions
'''

import time

from .common import ProblematicSymbols, keyexists, _clearexpired


def _keyexists(instance, db, key, create=True):
    tables = [f'{key}﹁bytes']
    initquery = f'INSERT INTO `{key}﹁bytes` VALUES(0)'
    return keyexists('bytes', tables, instance, db,
                     key, create, initquery=initquery)


def get(instance, key):
    'https://redis.io/commands/get'
    db = instance.router.connection(key)
    exists = _keyexists(instance, db, key, create=False)
    try:
        if not exists:
            return None
        res = db.execute(f'SELECT value FROM `{key}﹁bytes` LIMIT 1').fetchone()
        return res[0]
    finally:
        db.close()


def sēt(instance, key, value):
    'https://redis.io/commands/set'
    db = instance.router.connection(key)
    _keyexists(instance, db, key, create=True)
    try:
        db.execute(f'UPDATE `{key}﹁bytes` SET value=? LIMIT 1', (value, ))
        db.commit()
        _clearexpired(instance, db, key)
    finally:
        db.close()


def delete(instance, key):
    'https://redis.io/commands/del'
    db = instance.router.connection(key)
    try:
        ekey = key.replace('"', '""')
        tables = db.execute(
            'SELECT name FROM sqlite_master WHERE '
            f'type="table" AND name LIKE "{ekey}%"').fetchall()
        queries = ';\n'.join([f'DROP TABLE `{table[0]}`' for table in tables])
        db.executescript(queries)
        db.commit()
        if key in instance.knownkeys:
            del instance.knownkeys[key]
        _clearexpired(instance, db, key)
    finally:
        db.close()


def expire(instance, key, duration: int) -> int:
    'https://redis.io/commands/expire'
    db = instance.router.connection(key)
    if not _keyexists(instance, db, key, create=False):
        return 0
    try:
        expiretime = duration + time.time()
        db.execute(
            'INSERT OR REPLACE INTO `﹁expiredkeys﹁` (key, time) VALUES(?, ?)',
            (key, expiretime)
        )
        instance.keystoexpire[key] = expiretime
        return 1
    finally:
        db.close()


def persist(instance, key) -> int:
    'https://redis.io/commands/persist'
    db = instance.router.connection(key)
    if not _keyexists(instance, db, key, create=False):
        return 0
    if instance.checkexpired(key) != 'scheduled':
        return 0
    try:
        _clearexpired(instance, db, key)
        return 1
    finally:
        db.close()


def incr(instance, key, amount=1):
    'https://redis.io/commands/incr'
    db = instance.router.connection(key)
    _keyexists(instance, db, key, create=True)
    try:
        res = db.execute(f'SELECT value FROM `{key}﹁bytes` LIMIT 1').fetchone()
        try:
            val = int('0' if res is None else res[0])
        except ValueError:
            raise TypeError(f'value stored at {key} is not an integer')
        val += amount
        db.execute(f'UPDATE `{key}﹁bytes` SET value=? LIMIT 1', (val, ))
        db.commit()
        return val
    finally:
        db.close()


def incrby(instance, key, amount):
    'https://redis.io/commands/incrby'
    return incr(instance, key, amount)


def decr(instance, key, amount=1):
    'https://redis.io/commands/decr'
    db = instance.router.connection(key)
    _keyexists(instance, db, key, create=True)
    try:
        res = db.execute(f'SELECT value FROM `{key}﹁bytes` LIMIT 1').fetchone()
        try:
            val = int('0' if res is None else res[0])
        except ValueError:
            raise TypeError(f'value stored at {key} is not an integer')
        val -= amount
        db.execute(f'UPDATE `{key}﹁bytes` SET value=? LIMIT 1', (val, ))
        db.commit()
        return val
    finally:
        db.close()


def decrby(instance, key, amount):
    'https://redis.io/commands/decrby'
    return decr(instance, key, amount)


def flushdb(instance):
    'https://redis.io/commands/flushdb'
    instance.reset()


def keys(instance, pattern='*'):
    '''
    https://redis.io/commands/keys
    NOTE: there is no guarantee the keys returned are complete or precise, i.e.
    some existing keys might be missing and some retured keys might be deleted.
    For now we favor performance for this specific command, please use hkeys as
    a reliable alternative.
    '''
    allkeys = []
    pattern = pattern.replace('*', '%')
    for db in instance.router.allconnections():
        try:
            query = ('SELECT name FROM sqlite_master WHERE '
                     f'name LIKE "{pattern}" AND '
                     'name NOT LIKE "sqlite_%"')
            for record in db.execute(query):
                table = record[0]
                if table.startswith('﹁'):
                    continue
                key = table.split('﹁')[0]
                allkeys.append(ProblematicSymbols.restore(key))
        finally:
            db.close()
    return allkeys


def scan(instance, cursor=None, match='*'):
    '''
    https://redis.io/commands/scan
    currently is the same as `keys`
    '''

    if cursor not in (None, True):
        raise ValueError('invalid cursor %r' %cursor)

    if cursor is True:
        return []

    return True, keys(instance, match)



# -*- coding: utf-8 -*-

'''
common utilities
'''


def keyexists(keytype, tables, instance, db, key,
              create=True, columns=None, initquery=None):
    'checks if a key of given type exists, optionally creates one if not'

    wrongtypemsg = f'key `{key}` already exists with a different type'
    existingtype = instance.knownkeys.get(key)
    if existingtype:
        if existingtype != keytype:
            raise ValueError(wrongtypemsg)
        return True
    ekey = key.replace('"', '""')
    existing = db.execute('SELECT name FROM sqlite_master WHERE '
                          f'type="table" AND name LIKE "{ekey}%"').fetchall()
    existing = {record[0] for record in existing}
    if existing and existing != set(tables):
        raise ValueError(wrongtypemsg)

    if not existing:
        if not create:
            return False
        if not columns:
            columns = 'value TEXT'
        for table in tables:
            db.execute(f'CREATE TABLE `{table}` ({columns})')
            if initquery:
                db.execute(initquery)

    instance.knownkeys[key] = keytype
    return True


def _clearexpired(instance, db, key):
    if instance.checkexpired(key) in (True, 'scheduled'):
        db.execute('DELETE FROM `﹁expiredkeys﹁` WHERE key=?', (key,))
        db.commit()
        if key in instance.keystoexpire:
            del instance.keystoexpire[key]


class ProblematicSymbols:
    'remove symbols that could cause sql injection or use reserved words'

    FWD = {
        'sqlite_': 'sqlite﹁',
        '`': "﹁'﹁",
    }

    BWD = {v: k for k, v in FWD.items()}

    @staticmethod
    def remove(text):
        'remove symbols'
        for kw, rep in ProblematicSymbols.FWD.items():
            text = text.replace(kw, rep)
        return text

    @staticmethod
    def restore(text):
        'resotre symbols'
        for kw, rep in ProblematicSymbols.BWD.items():
            text = text.replace(kw, rep)
        return text



# -*- coding: utf-8 -*-

'''
router.py: provides the Router class to map keys to a corresponding database.
routing is configured by a spread factor
'''

import os
import sqlite3
import hashlib
import time
import uuid
from ilock import ILock


class DBConnection:
    'a database connection used within a pool'
    def __init__(self, fname):
        self.id = uuid.uuid1().hex
        self.free = True
        conn = sqlite3.connect(fname, check_same_thread=False)
        conn.execute('PRAGMA SYNCHRONOUS=OFF')
        conn.execute('PRAGMA CACHE_SIZE=500')
        query = (
            'CREATE TABLE IF NOT EXISTS `﹁expiredkeys﹁` '
            '(key TEXT PRIMARY KEY, time REAL)'
        )
        conn.execute(query)
        conn.commit()
        self._conn = conn

    def acquire(self):
        'acquires this connection'
        self.free = False

    def cursor(self):
        'sqlite3 cursor'
        return self._conn.cursor()

    def execute(self, *args, **kwargs):
        'sqlite3 execute'
        return self._conn.execute(*args, **kwargs)

    def executemany(self, *args, **kwargs):
        'sqlite3 executemany'
        return self._conn.executemany(*args, **kwargs)

    def executescript(self, *args, **kwargs):
        'sqlite3 executescript'
        return self._conn.executescript(*args, **kwargs)

    def commit(self):
        'sqlite3 commit'
        self._conn.commit()

    def close(self):
        'frees this connection'
        self.free = True


class Router:
    'main routing class, maps keys to a db based on a spread factor'
    EXT = 'clodssdb'

    def __init__(self, dbpath, factor=2, poolsize=5):
        '''
        dbpath: where to store the data files
        factor: partitioning factor, the higher it is, the more spread your
        data will be. this improves concurrency but also increases the number
        of open files. defaults to 2
        poolsize: pool size per data file, defaults to 5
        '''
        self.factor = factor
        self.dbpath = dbpath
        self.poolsize = poolsize
        self.connections = {}

    def _alldbs(self):
        return sorted([
            os.path.join(self.dbpath, f)[:-len(Router.EXT)-1]
            for f in os.listdir(self.dbpath)
            if f.endswith(Router.EXT)
        ])

    def _acquireconnection(self, db):
        'given an encoded db name, returns a connection'
        fname = os.path.join(self.dbpath, f'{db}.{Router.EXT}')
        with ILock(f'clodss-{db}-lock', timeout=5):
            conns = self.connections.get(db)
            if not conns:
                conns = [DBConnection(fname) for _ in range(self.poolsize)]
                self.connections[db] = conns

            for _ in range(1024):
                for conn in conns:
                    if not conn.free:
                        continue
                    conn.acquire()
                    return conn
                time.sleep(0.5)

    def reset(self):
        'clears the database and closes all connections'
        for db in self._alldbs():
            try:
                [conn.close() for conn in self.connections[db]]
            except KeyError:
                pass
            fname = os.path.join(self.dbpath, f'{db}.{Router.EXT}')
            os.unlink(fname)
        self.connections = {}

    def connection(self, key: bytes):
        'gets a new connection'
        db = hashlib.sha1(key.encode('utf-8')).hexdigest()[:self.factor]
        return self._acquireconnection(db)

    def allconnections(self, offset: int = 0):
        '''
        generator of connections to all available dbs
        offset: start offset in the alphabetically sorted list of dbs
        '''
        dbs = self._alldbs()[offset:]
        for db in dbs:
            yield self._acquireconnection(db)

    def poolstatus(self):
        'gets pooling status'
        return {k: [conn.free for conn in conns]
                for k, conns in self.connections.items()}



# -*- coding: utf-8 -*-

'''
clodss: list data-structure
'''

from .common import keyexists


def _listexists(instance, db, key, create=True):
    tables = [f'{key}﹁l', f'{key}﹁r']
    return keyexists('list', tables, instance, db, key, create)


def llen(instance, key) -> int:
    'https://redis.io/commands/llen'
    db = instance.router.connection(key)
    try:
        left = db.execute(f'SELECT COUNT(*) FROM `{key}﹁l`').fetchone()
        right = db.execute(f'SELECT COUNT(*) FROM `{key}﹁r`').fetchone()
        return left[0] + right[0]
    finally:
        db.close()


def rpush(instance, key, val):
    'https://redis.io/commands/rpush'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        db.execute(f'INSERT INTO `{key}﹁r` VALUES(?)', (val,))
        db.commit()
        return 1
    finally:
        db.close()


def lpush(instance, key, val):
    'https://redis.io/commands/lpush'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        db.execute(f'INSERT INTO `{key}﹁l` VALUES(?)', (val,))
        db.commit()
        return 1
    finally:
        db.close()


def rpop(instance, key):
    'https://redis.io/commands/rpop'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        table = f'{key}﹁r'
        res = db.execute(
            f'SELECT ROWID, value FROM `{table}` ORDER BY ROWID DESC LIMIT 1'
        ).fetchone()
        if not res:
            table = f'{key}﹁l'
            res = db.execute(f'''SELECT ROWID, value FROM `{table}`
                             ORDER BY ROWID ASC LIMIT 1''').fetchone()
            if not res:
                return None
        rowid, val = res
        db.execute(f'DELETE FROM `{table}` WHERE ROWID=?', (rowid,))
        db.commit()
        return val
    finally:
        db.close()


def lpop(instance, key):
    'https://redis.io/commands/lpop'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        table = f'{key}﹁l'
        res = db.execute(
            f'SELECT ROWID, value FROM `{table}` ORDER BY ROWID DESC LIMIT 1'
        ).fetchone()
        if not res:
            table = f'{key}﹁r'
            res = db.execute(f'''SELECT ROWID, value FROM `{table}`
                             ORDER BY ROWID ASC LIMIT 1''').fetchone()
            if not res:
                return None
        rowid, val = res
        db.execute(f'DELETE FROM `{table}` WHERE ROWID=?', (rowid,))
        db.commit()
        return val
    finally:
        db.close()


def _locateindex(db, key, index: int):
    'finds the table and offset for a given index'
    table = f'{key}﹁l'
    res = db.execute(f'SELECT COUNT(*) FROM `{table}`').fetchone()
    lcount = res[0]
    if index < lcount:
        order = 'DESC'
    else:
        index -= lcount
        table = f'{key}﹁r'
        order = 'ASC'
    return index, table, order


def _normalize_index(i: int, size: int, allowoverflow=True) -> int:
    'normalizes negative and oob indices'
    if i < 0:
        i += (abs(i)//size + 1) * size
        return i % size

    if i >= size and not allowoverflow:
        return None
    return min(i, size - 1)


def lindex(instance, key, index: int):
    'https://redis.io/commands/lindex'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        size = llen(instance, key)
        index = _normalize_index(index, size, False)
        if index is None:
            return None
        index, table, order = _locateindex(db, key, index)
        query = f'''SELECT value FROM `{table}`
                    ORDER BY ROWID {order} LIMIT 1 OFFSET {index}'''
        res = db.execute(query).fetchone()
        return res[0]
    finally:
        db.close()


def lset(instance, key, index: int, value):
    'https://redis.io/commands/lset'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        size = llen(instance, key)
        index = _normalize_index(index, size, False)
        if index is None:
            raise ValueError('out of bound index')
        index, table, order = _locateindex(db, key, index)
        query = f'''SELECT ROWID FROM `{table}`
                    ORDER BY ROWID {order} LIMIT 1 OFFSET {index}'''
        rowid = db.execute(query).fetchone()[0]
        db.execute(f'UPDATE `{table}` SET VALUE=? WHERE ROWID=?',
                   (value, rowid))
        db.commit()
    finally:
        db.close()


def lrange(instance, key, start: int, end: int):
    'https://redis.io/commands/lrange'
    db = instance.router.connection(key)
    cursor = db.cursor()
    cursor2 = db.cursor()
    _listexists(instance, db, key)
    try:
        size = llen(instance, key)
        start = _normalize_index(start, size)
        end = _normalize_index(end, size)

        if start > end:
            raise ValueError('end should be >= start')

        sindex, stable, sorder = _locateindex(db, key, start)
        eindex, etable, eorder = _locateindex(db, key, end)

        if stable == etable:
            n = eindex - sindex + 1
            query = f'''SELECT value FROM `{stable}`
                        ORDER BY ROWID {sorder} LIMIT {n} OFFSET {sindex}'''
            cursor.execute(query)
            return [record[0] for record in cursor]

        cursor.execute(f'''SELECT value FROM `{stable}`
                    ORDER BY ROWID {sorder} LIMIT -1 OFFSET {sindex}''')

        cursor2.execute(f'''SELECT value FROM `{etable}`
                    ORDER BY ROWID {eorder} LIMIT {eindex + 1} OFFSET 0''')
        return [record[0] for record in cursor] + [
            record[0] for record in cursor2]
    finally:
        cursor.close()
        cursor2.close()
        db.close()


def ltrim(instance, key, start: int, end: int) -> None:
    'https://redis.io/commands/ltrim'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        size = llen(instance, key)
        s, e = start, end
        start = _normalize_index(start, size)
        end = _normalize_index(end, size)

        truncate = s > e >= 0
        if truncate or start > end or start >= size:
            db.executescript('\n'.join([
                f'DELETE FROM `{key}﹁l`;',
                f'DELETE FROM `{key}﹁r`;'
                ]))
            db.commit()
            return None

        sindex, stable, sorder = _locateindex(db, key, start)
        eindex, etable, _ = _locateindex(db, key, end)

        if stable != etable:
            rlen = db.execute(f'SELECT COUNT(*) FROM `{stable}`').fetchone()[0]
            sindex = rlen - sindex
            eindex += 1
            db.execute(f'DELETE FROM `{stable}` LIMIT -1 OFFSET {sindex}')
            db.execute(f'DELETE FROM `{etable}` LIMIT -1 OFFSET {eindex}')
        else:
            if sorder == 'ASC':
                db.execute(f'DELETE FROM `{key}﹁l`')
                rsize = eindex - sindex + 1
                db.execute(f'DELETE FROM `{stable}` LIMIT {sindex} OFFSET 0')
                db.execute(f'DELETE FROM `{stable}` LIMIT -1 OFFSET {rsize}')
            else:
                db.execute(f'DELETE FROM `{key}﹁r`')
                rlen = db.execute(
                    f'SELECT COUNT(*) FROM `{stable}`').fetchone()[0]
                rsize = eindex - sindex + 1
                sindex = rlen - eindex - 1
                db.execute(f'DELETE FROM `{stable}` LIMIT {sindex} OFFSET 0')
                db.execute(f'DELETE FROM `{stable}` LIMIT -1 OFFSET {rsize}')
        db.commit()
        return None
    finally:
        db.close()


def lrem(instance, key, count: int, value) -> None:
    'https://redis.io/commands/lrem'
    db = instance.router.connection(key)
    _listexists(instance, db, key)
    try:
        if count == 0:
            db.execute(f'DELETE FROM `{key}﹁l` WHERE value=?', (value,))
            db.execute(f'DELETE FROM `{key}﹁r` WHERE value=?', (value,))
            db.commit()
            return None

        if count < 0:
            tables = [(f'{key}﹁r', 'DESC'), (f'{key}﹁l', 'ASC')]
            limit = -count
        else:
            tables = [(f'{key}﹁l', 'DESC'), (f'{key}﹁r', 'ASC')]
            limit = count

        nremoved = 0
        for table, order in tables:
            size = limit - nremoved
            res = db.execute(f'''DELETE FROM `{table}` WHERE value=?
                             ORDER BY ROWID {order} LIMIT {size}''',
                             (value, ))
            nremoved += res.rowcount
            if nremoved == limit:
                break
        db.commit()
        return None
    finally:
        db.close()


def linsert(instance, key, where, refvalue, value) -> int:
    'https://redis.io/commands/linsert'
    db = instance.router.connection(key)
    cursor = db.cursor()
    exists = _listexists(instance, db, key, False)
    try:
        if not exists:
            return 0

        size = llen(instance, key)
        where = where.lower()

        table = f'{key}﹁l'
        res = db.execute(f'''SELECT ROWID FROM `{table}` WHERE value=?
                         ORDER BY ROWID DESC LIMIT 1''',
                         (refvalue, )).fetchone()
        if res is not None:
            rowid = res[0]
            db.execute(f'INSERT INTO `{table}` VALUES("")')
            db.commit()
            op = '>' if where == 'before' else '>='
            cursor.execute(f'''SELECT ROWID, value FROM `{table}`
                           WHERE ROWID {op} {rowid}''')
            prv = value
            params = []
            for rowid, val in cursor:
                params.append((prv, rowid))
                prv = val
            db.executemany(f'UPDATE `{table}` SET value=? WHERE ROWID=?',
                           params)
            db.commit()
            return size + 1

        table = f'{key}﹁r'
        query = f'''SELECT ROWID FROM `{table}` WHERE value=?
                    ORDER BY ROWID ASC LIMIT 1'''
        res = db.execute(query, (refvalue, )).fetchone()
        if res is not None:
            rowid = res[0]
            db.execute(f'INSERT INTO `{table}` VALUES("")')
            db.commit()
            op = '>=' if where == 'before' else '>'
            cursor.execute(f'''SELECT ROWID, value FROM `{table}`
                           WHERE ROWID {op} {rowid}''')
            prv = value
            params = []
            for rowid, val in cursor:
                params.append((prv, rowid))
                prv = val
            db.executemany(f'UPDATE `{table}` SET value=? WHERE ROWID=?',
                           params)
            db.commit()
            return size + 1

        return size
    finally:
        cursor.close()
        db.close()
